#
#  Copyright: Stefan Doll (GNU License applies)
#
#  Module: fsm2vhdl
#
#  Function: Filter which creates a VHDL FSM from
#            a Tgif file
#
#  Revision History:
#  -----------------
#  V 1.0 :  6/5/96 :
#  V 1.1 : 15/2/97 : problem with gawk-3.0.0 replaced \'  with '
#

BEGIN \
{
statenum=0;
actnum=0;
decnum=0;
labelnum=0;
fsmnum=0;
wirenum=0;
def_name="";
fsm_clk="CLK'event and CLK='1'";
fsm_sync_reset="false";
fsm_rst="RST = '1'";
fsm_y0="";


fsm_object="none";
register_wire=0;
reg_line=0;
reg_black_text=0;
reg_red_text=0;
line_points=0;
line=0;
y0line=0;
register_y0=0;
chk_y0=0;
}

# parse tgif file


# check object label
/^\t\"###/ \
  {
   split($1,parts,"\"");
   fsm_object=tolower(substr(parts[2],4,length(parts[2])-3));
   #*print "fsm_object = " fsm_object;
  }

# a red line indicates a label-object
/^poly\('red'/  \
  {
   fsm_object="label";
  }


# register lines
/]/ \
  {
   if(reg_line)
     {
      for(i=1;i<=4;i++)
        {
          lp[line_points]=$(i);
	  line_points++;
        }
      lp[line_points-1]=substr($4,1,length($4)-1);
      reg_line=0;
      FS=" ";
     }
  }
/^poly\('/  \
  {
   reg_line=1;
   FS=",";
  }


# register black text
/^\t\"/ \
  {
   if(reg_black_text)
     {
      black_text[line]=substr($0,3,length($0)-4);
      #print "line=" line;
      #print black_text[line];
      line++;
     }
  }
/\"]\)/   \
  {
   if(reg_black_text) reg_black_text=0;
  }
/^text\('black'/  \
  {
   reg_black_text=1;
  }


# register Y0
/^\t\"/ \
  {
   if(chk_y0 && match($0,"Y0:")) register_y0=1;
   if(register_y0)
     {
      y0[y0line]=substr($0,3,length($0)-4);
      gsub("Y0:","",y0[y0line]);
      y0line++;
     }
  }
/\"]\)/   \
  {
   chk_y0=0;
   register_y0=0;
  }
/^text\('black'/  \
  {
   chk_y0=1;
  }


# register red text
/^\t\"/ \
  {
   if(reg_red_text)
     {
      red_text[line]=substr($0,3,length($0)-4) " ";
      # eliminate spare `"' when on last line
      if(substr($0,length($0))==")") red_text[line]=substr($0,3,length($0)-5);
      gsub("\\" "\\", "", red_text[line])
      line++;
     }
  }
/\"]\)/   \
  {
   if(reg_red_text) reg_red_text=0;
  }
/^text\('red'/  \
  {
   reg_red_text=1;
  }


# register diagonal points of a box
/^box\('/ \
  {
   split($0,parts,",");
   bp_x[0]=parts[2];
   bp_y[0]=parts[3];
   bp_x[1]=parts[4];
   bp_y[1]=parts[5];
   #print "#### " bp_x[0] "." bp_y[0] "/" bp_x[1] "." bp_y[1]
  }


#
# collecting data about wiring
#
/^poly\('blue'/ \
  {
    register_wire=1;
    split($0, def_string, ",");
    points[wirenum]=def_string[2];
    FS=",";
  }

/]/ \
   {
    if(register_wire)
     {
       last_tuppel="false";
       for (n=0; n<points[wirenum]; n++)
        {
 	wx[wirenum, n]=$(2*n+1)+1-1;
 	wy[wirenum, n]=$(2*n+2);
        };
       wy[wirenum, n-1]= \
       substr(wy[wirenum, n-1], 1, length(wy[wirenum, n-1])-1);
       # ensure all points become integers
       for (n=0; n<=points[wirenum]; n++)
        {
 	wx[wirenum, n]++; wx[wirenum, n]--;
 	wy[wirenum, n]++; wy[wirenum, n]--;
        };
       wire_name[wirenum]="#none#";
       wire_type[wirenum]="#none#";
       wire_is_subnet[wirenum]="false";
       id[wirenum]=wirenum;
       register_wire=0;
       wirenum++; 
       FS=" ";
    }
  }



# end of a group, assemble object
/^])\./ \
  {
   if(lp[1] < lp[5]) # first line indicates input
     {
      in_x1=lp[0];
      in_y1=lp[1];
      in_x2=lp[2];
      in_y2=lp[3];
      out_x1=lp[4];
      out_y1=lp[5];
      out_x2=lp[6];
      out_y2=lp[7];
     }
   if(lp[1] > lp[5]) # second line indicates input
     {
      in_x1=lp[4];
      in_y1=lp[5];
      in_x2=lp[6];
      in_y2=lp[7];
      out_x1=lp[0];
      out_y1=lp[1];
      out_x2=lp[2];
      out_y2=lp[3];
     }

   for(l=0;l<line;l++) gsub("\\" "\\", "", black_text[l]);
   black_text[line-1] = substr(black_text[line-1],1,     \
                        length(black_text[line-1])-2);
   if(fsm_object=="state")
     {
      state_in_x1[statenum]=in_x1;
      state_in_y1[statenum]=in_y1;
      state_in_x2[statenum]=in_x2;
      state_in_y2[statenum]=in_y2;
      state_out_x1[statenum]=out_x1;
      state_out_y1[statenum]=out_y1;
      state_out_x2[statenum]=out_x2;
      state_out_y2[statenum]=out_y2;
      state_name[statenum]=black_text[0];
      state_rline_x0[statenum]=bp_x[1];
      state_rline_y0[statenum]=bp_y[0];
      state_rline_x1[statenum]=bp_x[1];
      state_rline_y1[statenum]=bp_y[1];
      statenum++;
     }

   if(fsm_object=="action")
     {
      act_in_x1[actnum]=in_x1;
      act_in_y1[actnum]=in_y1;
      act_in_x2[actnum]=in_x2;
      act_in_y2[actnum]=in_y2;
      act_out_x1[actnum]=out_x1;
      act_out_y1[actnum]=out_y1;
      act_out_x2[actnum]=out_x2;
      act_out_y2[actnum]=out_y2;
      for(l=0;l<line;l++) act_vhdl[actnum, l]=black_text[l];
      act_vhdl_lines[actnum]=line;
      actnum++;
     }

   if(fsm_object=="decision")
     {
      dec_in_x1[decnum]=in_x1;
      dec_in_y1[decnum]=in_y1;
      dec_in_x2[decnum]=in_x2;
      dec_in_y2[decnum]=in_y2;
      dec_out_x1[decnum]=out_x1;
      dec_out_y1[decnum]=out_y1;
      dec_out_x2[decnum]=out_x2;
      dec_out_y2[decnum]=out_y2;
      for(l=0;l<line;l++) dec_expr[decnum]=dec_expr[decnum] black_text[l];
      decnum++;
     }

   # info for definition object
   for(l=0;l<line;l++)
     {
      if(substr(black_text[l],1,4)=="FSM=")
        {fsm_string=substr(black_text[l],5,length(black_text[l])-6);}
      if(match(black_text[l],"CLK:"))
        {clk_string=substr(black_text[l],5,length(black_text[l])-6);}
      if(substr(black_text[l],1,6)=="RESET:")
        {rst_string=substr(black_text[l],7,length(black_text[l])-8);}
      if(substr(black_text[l],1,10)=="sync_reset")
        {
         sync_reset="false";
         if(match(tolower(black_text[l]),"true")) sync_reset="true";
        }
     }

   if(fsm_object=="def")
     {
      def_name=fsm_string;
      fsm_clk=clk_string;
      fsm_sync_reset=sync_reset;
      fsm_rst=rst_string;
      for(l=0;l<y0line;l++) gsub("\\" "\\", "", y0[l]);
      y0[y0line-1] = substr(y0[y0line-1],1,length(y0[y0line-1])-2);
      for(l=0;l<y0line;l++) fsm_y0=fsm_y0 "\t" y0[l] "\n";
      y0line=0;
     }

   if(fsm_object=="label")
     {
      if(   ((lp[0]-lp[2])**2 + (lp[1]-lp[3])**2)   \
          > ((lp[4]-lp[6])**2 + (lp[5]-lp[7])**2) )
        {
	 # first line is the wire
         lbl_x1[labelnum] = lp[0]; lbl_y1[labelnum] = lp[1];
         lbl_x2[labelnum] = lp[2]; lbl_y2[labelnum] = lp[3]
        }
       else
        {
	 # second line is the wire
         lbl_x1[labelnum] = lp[4]; lbl_y1[labelnum] = lp[5];
         lbl_x2[labelnum] = lp[6]; lbl_y2[labelnum] = lp[7];
        }

      #*# additionally register the longer red line as a wire
      #*wx[wirenum, 0]=lbl_x1[labelnum];
      #*wy[wirenum, 0]=lbl_y1[labelnum];
      #*wx[wirenum, 1]=lbl_x2[labelnum];
      #*wy[wirenum, 1]=lbl_y2[labelnum];
      #*points[wirenum]=2;
      #*wire_name[wirenum]="#none#";
      #*wire_type[wirenum]="#none#";
      #*wire_is_subnet[wirenum]="false";
      #*id[wirenum]=wirenum;
      #*wirenum++; 
 
      #register label expression
      for(l=0;l<line;l++) lbl_expr[labelnum]=lbl_expr[labelnum] red_text[l];
      labelnum++;
     }

   if(fsm_object=="start")
     {
      start_x[fsmnum]=lp[0];
      start_y[fsmnum]=lp[1];
      if(lp[2]<lp[0]) {
        start_x[fsmnum]=lp[2];
        start_y[fsmnum]=lp[3];
       }
      gsub("FSM=", "",black_text[0])
      fsm_name[fsmnum]=black_text[0];
      fsmnum++;
     }



   fsm_object="none";
   line_points=0;
   line=0;
  }


# processing input data


# grouping wires to nets




# identifying whether a point
# is on a horizantal or vertical line

function point_on_line(ptx, pty, lx0, ly0, lx1, ly1, \
		      y0, y1, x0, x1, on_line)
 {
	on_line="false";
	#print "###";
	#print "lx0 " lx0 "lx1 " lx1 "ptx" ptx;

	# point on vertical line
	if (lx0==lx1 && ptx==lx0)
	  {
	   y0=ly0; y1=ly1;
	   if (ly0>ly1) {y0=ly1; y1=ly0}
           if(pty>=y0 && pty<=y1) on_line="true";
	  }

	# point on horizontal line
	if (ly0==ly1 && pty==lx0)
	  {
	   x0=lx0; x1=lx1;
	   if (lx0>lx1) {x0=lx1; x1=lx0}
           if(ptx>=x0 && ptx<=x1) on_line="true";
	  }

	return on_line;
 }



# identifying whether a point
# is on a horizontal or vertical segment
#
function point_on_seg(ptx, pty, wire, pt, \
		      y0, y1, on_seg)
 {
	on_seg="false";
	# point on vertical segment
	if (wy[wire,s]>wy[wire,s+1])
  	   {y0=wy[wire,pt+1]; y1=wy[wire,pt]}
	 else
 	   {y0=wy[wire,pt]; y1=wy[wire,pt+1]};
	if (ptx==wx[wire,pt] && wx[wire,pt]==wx[wire,pt+1] && \
	    pty>=y0 && pty<=y1)
	    on_seg="true";

	# point on horizontal segment
	if (wx[wire,pt]>wx[wire,pt+1])
		{x0=wx[wire,pt+1]; x1=wx[wire,pt]}
	  else
		{x0=wx[wire,pt]; x1=wx[wire,pt+1]};
	if (pty==wy[wire,pt] && wy[wire,pt]==wy[wire,pt+1] && \
	    ptx>=x0 && ptx<=x1)
	    on_seg="true";

	return on_seg;
 }



# identifying whether a line 
# is on a horizantal or vertical segment
#
function line_on_seg(lx0, ly0, lx1, ly1, wire, pt, \
		      a, b, w0, w1, on_seg)
 {

  	on_seg="false";

	# overlapping vertical segments
	if (lx0==lx1 && wx[wire,pt]==wx[wire,pt+1] && \
	    lx0==wx[wire,pt])
	  {
 	    if (ly0>ly1)
  		{a=ly1; b=ly0}
  	       else
  		{a=ly0; b=ly1};
  	    if (wy[wire,pt]>wy[wire,pt+1])
  		{w0=wy[wire,pt+1]; w1=wy[wire,pt]}
  	       else
  		{w0=wy[wire,pt]; w1=wy[wire,pt+1];}

	    if ( (w0>=a && w0<=b) || \
  	         (w1>=a && w1<=b) || \
	         (a>=w0 && a<=w1) )
  	         on_seg="true";
 	   };

  	# overlapping horizontal segments
  	if (ly0==ly1 && wy[wire,pt]==wy[wire,pt+1] && \
  	    ly0==wy[wire,pt])
  	  {
  	    if (lx0>lx1)
  		{a=lx1; b=lx0}
  	       else
  		{a=lx0; b=lx1};
  	    if (wx[wire,pt]>wx[wire,pt+1])
  		{w0=wx[wire,pt+1]; w1=wx[wire,pt]}
  	       else
  		{w0=wx[wire,pt]; w1=wx[wire,pt+1];}

	    if ( (w0>=a && w0<=b) || \
  	         (w1>=a && w1<=b) || \
	         (a>=w0 && a<=w1) )
  	         on_seg="true";
  	  };

	return on_seg;

}




END \
{
   # for all wires w1
   # (last wire will not be compared since there 
   #  will be nothing left to compare it with)
   for (w1=0; w1<wirenum-1; w1++)
    {
       # for all wire segments of w1
       for (s=0; s<points[w1]-1; s++)
        {
 	# for all other wires w2
         for (w2=w1+1; w2<=wirenum-1; w2++)
 	  {
 	   wires_connected="false";
  
 	   # matching endpoints 
	   #
	   # -first and first
 	   w1x=wx[w1,0]; w1y=wy[w1,0];
  	   w2x=wx[w2,0]; w2y=wy[w2,0];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -first and last
	   w2x=wx[w2,points[w2]-1]; w2y=wy[w2,points[w2]-1];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -last and last
	   w1x=wx[w1,s+1]; w1y=wy[w1,s+1];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -last and first
  	   w2x=wx[w2,0]; w2y=wy[w2,0];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";

	   # check whether w2's endpoints are on current w1 segment
	   #
	   # first point on segment
	   if (point_on_seg(wx[w2,0], wy[w2,0], w1, s)=="true")
		wires_connected="true";
	   # last point on segment
	   if (point_on_seg(wx[w2,points[w2]-1], \
			    wy[w2,points[w2]-1], w1, s) == "true")
		wires_connected="true";

  
  	   # for all segments of w2
             for (t=0; t<=points[w2]-2; t++)
  	      {

		# check whether w1's endpoints are on current w2 segment
		# first on segment
		if (point_on_seg(wx[w1,0], wy[w1,0], w2, t) == "true")
 		   wires_connected= "true";
		# last on segment
		if (point_on_seg(wx[w1,points[w1]-1], \
		 		 wy[w1,points[w1]-1], w2, t) == "true")
		   wires_connected= "true";
   
 		# overlapping segments
		if (line_on_seg(wx[w2,t],wy[w2,t], wx[w2,t+1],wy[w2,t+1], w1,s) \
		    == "true") wires_connected= "true";
		
  	      }

 	    # group together
 	    if (wires_connected=="true")
	      { 
		replace_id=id[w2];
		for (v=0;v<wirenum;v++) if(id[v]==replace_id) id[v]=id[w1];
	      } 
 	  }
        }
    }	
 }



# determine the net to which an object pin is connected

function find_net_id(ox1,oy1, ox2,oy2, \
		     w1, connected_net)
  {
   # object inputs/outputs are stored as wires as well
   # => there must be a wire which has an input/output point
   #    as a starting point
   connected_net="unconnected";
   for (w1=0; w1<wirenum-1; w1++)
     {
      if( (ox1==wx[w1,0] && oy1==wy[w1,0]) || (ox2==wx[w1,0] && oy2==wy[w1,0]) )
        {
         connected_net=id[w1];
         break;
        }
     }
   return connected_net;
  }

function find_lbl_net_id(lx,ly, \
		         w1, connected_net)
  {
   # check for wire endpoints which are connected to the label
   connected_net="unconnected";
   for (w1=0; w1<wirenum-1; w1++)
     {
      if( (lx==wx[w1,0] && ly==wy[w1,0]) || \
          (lx==wx[w1,points[w1]-1] && ly==wy[w1,points[w1]-1]) )
        {
         connected_net=id[w1];
         break;
        }
     }
   return connected_net;
  }

END {

# find nets which are connected to state symbols

for(s=0;s<statenum;s++)
  {
   state_in_net[s]= \
         find_net_id(state_in_x1[s],state_in_y1[s], \
	             state_in_x2[s],state_in_y2[s]);
   state_out_net[s]= \
         find_net_id(state_out_x1[s],state_out_y1[s], \
	             state_out_x2[s],state_out_y2[s]);
  }


# find nets which are connected to action symbols

for(a=0;a<actnum;a++)
  {
   act_in_net[a]= \
         find_net_id(act_in_x1[a],act_in_y1[a], \
	             act_in_x2[a],act_in_y2[a]);
   act_out_net[a]= \
         find_net_id(act_out_x1[a],act_out_y1[a], \
	             act_out_x2[a],act_out_y2[a]);
  }


# find nets which are connected to decision symbols

for(d=0;d<decnum;d++)
  {
   dec_in_net[d]= \
         find_net_id(dec_in_x1[d],dec_in_y1[d], \
	             dec_in_x2[d],dec_in_y2[d]);
   dec_out_net[d]= \
         find_net_id(dec_out_x1[d],dec_out_y1[d], \
	             dec_out_x2[d],dec_out_y2[d]);
  }


# determine the net id for each label

for(l=0;l<labelnum;l++)
  {
   # look for nets which are connected to one endpoint
   lbl_net1[l]= find_lbl_net_id(lbl_x1[l],lbl_y1[l]);
   # look for nets which are connected to the other endpoint
   lbl_net2[l]= find_lbl_net_id(lbl_x2[l],lbl_y2[l]);
  }


# find the reset state
for(s=0;s<statenum;s++)
  {
  #print "state " s;
  if(point_on_line(start_x[0], start_y[0], \
		    state_rline_x0[s], state_rline_y0[s], \
                    state_rline_x1[s], state_rline_y1[s])=="true")
       reset_state=state_name[s];
  }

}



# generate VHDL
# --------------

# identifies all state transitions of a state 
#
function gen_trans_proc(net_id, \
		     a,d,s,l,m,first_label)
 {

   for(a=0;a<actnum;a++)
     {
      if(act_in_net[a]==net_id) # connected to an action block
        {
	 # continue looking from action block output
	 #*print "connected to action " a;          
	 gen_trans_proc(act_out_net[a]);
         return; # no need to continue looking for the old net
        } 
     }
    
   for(d=0;d<decnum;d++)
     {
      if(dec_in_net[d]==net_id) # connected to a decision block
        {
	 for(l=0;l<labelnum;l++)
	   {
	    #*print "connected to decision " d;          
	    if((lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d]) \
	       && match(tolower(lbl_expr[l]),"else") )
	      {
		first_label="true";
		# this is a decision node with an else clause
	 	for(m=0;m<labelnum;m++)
		   {
		    if((lbl_net1[m]==dec_out_net[d] || \
		        lbl_net2[m]==dec_out_net[d]) && m != l)
		      {
		       # build the condition which follows this branch
		       printf "\t\t";
		       if(first_label=="false") printf "els";
		       first_label="false";
		       print "if " dec_expr[d] " " lbl_expr[m] " then";

		       # find the actions which occur in here
		       # (nested IFs + state assignments)
	
		       if(lbl_net1[m]==dec_out_net[d])
			  gen_trans_proc(lbl_net2[m]);
		       if(lbl_net2[m]==dec_out_net[d]) 
			  gen_trans_proc(lbl_net1[m]);

		      }
		   }
	        # else clause
	        print "\t\telse";
	        # find the actions which occur in here
	        # (nested IFs + state assignments)
	        if(lbl_net1[l]==dec_out_net[d]) gen_trans_proc(lbl_net2[l]);
	        if(lbl_net2[l]==dec_out_net[d]) gen_trans_proc(lbl_net1[l]);

	        # end of the IF statement
	        print "\t\tend if;";
	     
	        return; # finished this level
	      }
	   }
	 for(l=0;l<labelnum;l++)
	   {
	    if(lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d])
	      {
	       #*print "connected to label " l;          
	       # build the condition which follows this branch
	       printf "\t\tif " dec_expr[d] " " lbl_expr[l];
               print " then";

	       # find the actions which occur in here
	       # (nested IFs + state assignments)

	       if(lbl_net1[l]==dec_out_net[d]) gen_trans_proc(lbl_net2[l]);
	       if(lbl_net2[l]==dec_out_net[d]) gen_trans_proc(lbl_net1[l]);

	       # end of the IF statement
               print "\t\tend if;";
	      }
	   }
         return; # no need to continue looking for the old net
        } 
     }

   for(s=0;s<statenum;s++)
     {
      if(state_in_net[s]==net_id) # found end of branch
        {
         print "\t\t   vgi_state <= " state_name[s] ";";
         return;
        }
     }

   return;
}


# identifies all outputs of a state (Moore or Mealy) 
#
function gen_outp_proc(net_id, \
		     a,d,s,l,m,first_label)
 {

   for(a=0;a<actnum;a++)
     {
      if(act_in_net[a]==net_id) # connected to an action block
        {
	 # execute action
         for(l=0;l<act_vhdl_lines[a];l++)
	   print "\t\t" act_vhdl[a, l];
	 # continue looking from action block output
	 #*print "connected to action " a;          
	 gen_outp_proc(act_out_net[a]);
         return; # no need to continue looking for the old net
        } 
     }
    
   for(d=0;d<decnum;d++)
     {
      if(dec_in_net[d]==net_id) # connected to a decision block
        {
	 #*print "connected to decision " d;          
	 for(l=0;l<labelnum;l++)
	   {
	    if((lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d]) \
	       && match(tolower(lbl_expr[l]),"else") )
	      {
		first_label="true";
		# this is a decision node with an else clause
	 	for(m=0;m<labelnum;m++)
		   {
		    if(lbl_net1[m]==dec_out_net[d] || \
		       lbl_net2[m]==dec_out_net[d] && m != l)
		      {
		       # build the condition which follows this branch
		       printf "\t\t";
		       if(first_label=="false") printf "els";
		       first_label="false";
		       print "if " dec_expr[d] " " lbl_expr[m] " then";

		       # find the actions which occur in here
		       # (nested IFs + state assignments)
	
		       if(lbl_net1[m]==dec_out_net[d])
			  gen_outp_proc(lbl_net2[m]);
		       if(lbl_net2[m]==dec_out_net[d]) 
			  gen_outp_proc(lbl_net1[m]);

		      }
		   }
	        # else clause
	        print "\t\telse";
	        # find the actions which occur in here
	        # (nested IFs + state assignments)
	        if(lbl_net1[l]==dec_out_net[d]) gen_outp_proc(lbl_net2[l]);
	        if(lbl_net2[l]==dec_out_net[d]) gen_outp_proc(lbl_net1[l]);

	        # end of the IF statement
	        print "\t\tend if;";
	     
	        return; # finished this level
	      }
	   }
	 for(l=0;l<labelnum;l++)
	   {
	    if(lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d])
	      {
	       #*print "connected to label " l;          
	       # build the condition which follows this branch
	       printf "\t\tif " dec_expr[d] " " lbl_expr[l];
               print " then";

	       # find the actions which occur in here
	       # (nested IFs + state assignments)

	       if(lbl_net1[l]==dec_out_net[d]) gen_outp_proc(lbl_net2[l]);
	       if(lbl_net2[l]==dec_out_net[d]) gen_outp_proc(lbl_net1[l]);

	       # end of the IF statement
               print "\t\tend if;";
	      }
	   }
         return; # no need to continue looking for the old net
        } 
     }

   for(s=0;s<statenum;s++)
     {
      if(state_in_net[s]==net_id) # found end of branch
        {
         return;
        }
     }

   return;
}



END \
{

  print "";
  print "--";
  print "-- FSM generated by VGI - fsm2vhdl";
  system("echo \"-- at\" `date` ");
  print "--";
  print "";
  print "";

  # extract state type

  printf "  type " fsm_name[0] "_type is (";
  for(s=0;s<statenum;s++)
    {
     printf state_name[s];
     if(s!=statenum-1) printf ", ";
    }
  print ");"
  print "";
  print "  signal " fsm_name[0] "_state: " fsm_name[0] "_type;";
  print "";
  print "";
  print "";
  print fsm_name[0] "_transition: process(CLK, RST)";
  print "";
  print "";
  print "begin";
  print "";
  print "   if " fsm_rst " then";
  print "";
  print "\t" fsm_name[0] "_state <= " reset_state ";";
  print "";
  print "   elsif " fsm_clk " then";
  print "";
  print "";
  print "";
  print "      case " fsm_name[0] "_state is";

  for(s=0;s<statenum;s++)
    {
     print "";
     printf "         when ";
     printf state_name[s];
     print " =>";
     print "";
     gen_trans_proc(state_out_net[s]);
    }
  print "";

  print "      end case;";

  print "";
  print "   end if;";
  print "";
  print "end process;";


  # generate the output process

  print "";
  print "";
  print fsm_name[0] "_output: process()";
  print "";
  print "";
  print "";
  print "begin";
  print "";
  print fsm_y0;
  print "";
  print "      case " fsm_name[0] "_state is";

  for(s=0;s<statenum;s++)
    {
     print "";
     printf "         when ";
     printf state_name[s];
     print " =>";
     print "";
     gen_outp_proc(state_out_net[s]);
    }
  print "";

  print "      end case;";

  print "";
  print "end process;";


}


END {
debug="false";
if(debug=="true")
 {
     # print Moore states

     # print the action which is connected with the current state
     for(a=0;a<actnum;a++)
       if(state_out_net[s]==act_in_net[a])
         {
          for(l=0;l<act_vhdl_lines[a];l++) 
              print "\t    " act_vhdl[a,l];
         }
     print "";

     for(d=0;d<decnum;d++) 
       {
        printf "\t    if " dec_expr[d];
        print " then";
        print "\t    ";
        print "\t    end if";
       }
  for(s=0;s<statenum;s++) 
   {
    print state_name[s] ":";
    print "in=" state_in_net[s];
    print "out=" state_out_net[s];
    print state_in_x1[s] ", " state_in_y1[s] " / " \
          state_in_x2[s] ", " state_in_y2[s];
    print state_out_x1[s] ", " state_out_y1[s] " / " \
          state_out_x2[s] ", " state_out_y2[s];

   }
  for(l=0;l<labelnum;l++) 
   {
    print lbl_expr[l] ":";
    print "net1=" lbl_net1[l];
    print "net2=" lbl_net2[l];
    print lbl_x1[l] ", " lbl_y1[l] " / " \
          lbl_x2[l] ", " lbl_y2[l];
   }
  for(a=0;a<actnum;a++) 
   {
    print "action " a ":";
    print "in=" act_in_net[a];
    print "out=" act_out_net[a];
   }
  for(d=0;d<decnum;d++) 
   {
    print "decision " d ":";
    print "in=" dec_in_net[d];
    print "out=" dec_out_net[d];
   }
		 for (n=0; n < wirenum; n++)
		 { 
			print "wirenr: " n " name: " wire_name[n];
			print "id: " id[n] " subnet=" wire_is_subnet[n];
		   print "points: " points[n];
		   print "id: " id[n];
		   for (m=0; m < points[n]; m++)
		     {
			print wx[n, m] "," wy[n, m];
		     }
		 }

 }

}
